#include <opencv2/core.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>

#include <string>
#include <iostream>

#include <nvbufsurface.h>
#include <nvbufsurftransform.h>


void configureNvBufSurfTransform();

// Create a surface with a single image of size sz and format f
void createNvBufSurface(NvBufSurface*& s, cv::Size sz, NvBufSurfaceColorFormat f);
// Transform contents of src to dst with provided algorithm
void transformNvBufSurface(NvBufSurface* src, NvBufSurface* dst, NvBufSurfTransform_Inter algorithm);
// memcpy the contents of a NvBufSurface to an OpenCV mat
// Preconditions: The mat and surface must have the same bytesPerPx and dimensions
void copyNvBufSurfaceToMat(cv::Mat& mat, NvBufSurface* surface);
// memcpy the contents of an OpenCV mat to a NvBufSurface
// It is expected that the layouts are contiguous (pitch padding is allowed) and that channels are interleaved in memory, that the Mat and surface have the same dimensions and channels
void copyMatToNvBufSurface(cv::Mat mat, NvBufSurface* surface);


int main(int argc, char** argv) {
	configureNvBufSurfTransform();

	// Load sample image
	cv::Mat BGR = cv::imread("pattern.png", /*cv::IMREAD_GRAYSCALE*/ cv::IMREAD_COLOR);
	cv::Mat a;
	cv::cvtColor(BGR, a, cv::COLOR_BGR2RGBA);

	// Create 2 surfaces. Then load the sample image into the first
	NvBufSurface* inputSurface, * scaledSurface;
	cv::Size downscaledSize = a.size() / 2;
	createNvBufSurface(inputSurface, a.size(), NVBUF_COLOR_FORMAT_RGBA);
	createNvBufSurface(scaledSurface, downscaledSize, NVBUF_COLOR_FORMAT_RGBA);
	copyMatToNvBufSurface(a, inputSurface);

	// Transform contents from first surface to second, using different algorithms
	std::vector<NvBufSurfTransform_Inter> algorithms_to_try{
		NvBufSurfTransform_Inter::NvBufSurfTransformInter_Algo1,
		NvBufSurfTransform_Inter::NvBufSurfTransformInter_Algo2,
		NvBufSurfTransform_Inter::NvBufSurfTransformInter_Algo4,
		NvBufSurfTransform_Inter::NvBufSurfTransformInter_Algo3,
		NvBufSurfTransform_Inter::NvBufSurfTransformInter_Bilinear,
		NvBufSurfTransform_Inter::NvBufSurfTransformInter_Nearest
	};

	std::vector<cv::Mat> results;
	for (int algo = 0; algo < algorithms_to_try.size(); algo++) {
		transformNvBufSurface(inputSurface, scaledSurface, algorithms_to_try[algo]);
		cv::Mat result(downscaledSize, CV_8UC4);
		copyNvBufSurfaceToMat(result, scaledSurface);
		
		cv::imwrite(std::string("result") + std::to_string(algo) + ".png", result);
		results.push_back(result);
	}
	
	// uncomment to intentionally create difference, to check the code
	// results[2].at<cv::Vec4i>(10, 10) = { 14, 16, 34, 2 };

	// Check that the contents really are the same
	for (auto it = results.begin() + 1; it != results.end(); ++it) {
		std::vector<cv::Mat> channels_first, channels_current;
		cv::split(*it, channels_current);
		cv::split(results[0], channels_first);

		for (int channel = 0; channel < it->channels(); channel++) {

			cv::Mat diff = channels_first[channel] != channels_current[channel];
			bool eq = cv::countNonZero(diff) == 0;
			if (eq) std::cout << "No difference in byte values for channel" << std::endl;
			else std::cout << "Byte values different" << std::endl;
		}
		
	}

	return 0;
}


#define NVBUF_ERROR_COUT_NOTICE(name, status) {if(status!=0){std::cout << name " failed with" << std::to_string(status) << std::endl;}}
void createNvBufSurface(NvBufSurface*& s, cv::Size sz, NvBufSurfaceColorFormat f) {

	NvBufSurfaceCreateParams p;

	memset(&p, 0, sizeof p);
	p.colorFormat = f;
	p.width = sz.width;
	p.height = sz.height;
	p.memType = NVBUF_MEM_SURFACE_ARRAY;
	p.layout = NVBUF_LAYOUT_PITCH;

	int nvbufState;
	nvbufState = NvBufSurfaceCreate(&s, 1, &p);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceCreate", nvbufState);

	s->numFilled = 1;
}

void configureNvBufSurfTransform() {
	NvBufSurfTransformConfigParams sp;
	memset(&sp, 0, sizeof(NvBufSurfTransformConfigParams));
	sp.compute_mode = NvBufSurfTransformCompute_VIC;

	auto status = NvBufSurfTransformSetSessionParams(&sp);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfTransformSetSessionParams", status);
}

void transformNvBufSurface(NvBufSurface* src, NvBufSurface* dst, NvBufSurfTransform_Inter algorithm) {
	NvBufSurfTransformParams tp;

	memset(&tp, 0, sizeof tp);

	NvBufSurfTransformRect tr_input;
	tr_input.height = src->surfaceList[0].height;
	tr_input.width = src->surfaceList[0].width;
	tr_input.top = tr_input.left = 0;
	tp.src_rect = &tr_input;

	NvBufSurfTransformRect tr_intermediate;
	tr_intermediate.height = dst->surfaceList[0].height;
	tr_intermediate.width = dst->surfaceList[0].width;
	tr_intermediate.top = tr_intermediate.left = 0;
	tp.dst_rect = &tr_intermediate;

	tp.transform_flip = NvBufSurfTransform_None;
	tp.transform_filter = algorithm;

	auto status = NvBufSurfTransform(src, dst, &tp);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfTransform", status);
}

void copyNvBufSurfaceToMat(cv::Mat& mat, NvBufSurface* surface) {
	auto status = NvBufSurfaceMap(surface, -1, -1, NVBUF_MAP_READ);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceMap", status);
	status = NvBufSurfaceSyncForCpu(surface, -1, -1);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceSyncForCpu", status);

	cv::Size sz(surface->surfaceList[0].width, surface->surfaceList[0].height);

	int matFormat = mat.type();
	cv::Mat surfaceWrapperMat(sz, matFormat, surface->surfaceList[0].mappedAddr.addr[0], surface->surfaceList[0].pitch);

	// Copy every row
	for (int i = 0; i < sz.height; i++) {
		memcpy(mat.ptr(i), surfaceWrapperMat.ptr(i), sz.width * surface->surfaceList[0].planeParams.bytesPerPix[0]);
	}

	status = NvBufSurfaceUnMap(surface, -1, -1);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceUnMap", status);
}

void copyMatToNvBufSurface(cv::Mat mat, NvBufSurface* surface) {
	auto status = NvBufSurfaceMap(surface, -1, -1, NVBUF_MAP_WRITE);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceMap", status);
	status = NvBufSurfaceSyncForCpu(surface, -1, -1);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceSyncForCpu", status);

	cv::Size sz{ surface->surfaceList[0].width, surface->surfaceList[0].height };

	int matFormat = mat.type();
	cv::Mat surfaceWrapperMat(sz, matFormat, surface->surfaceList[0].mappedAddr.addr[0], surface->surfaceList[0].pitch);
	// Copy each row, to make sure it doesn't allocate a new mat
	for (int i = 0; i < surfaceWrapperMat.rows; i++) {
		memcpy(surfaceWrapperMat.ptr(i), mat.ptr(i), sz.width * surface->surfaceList[0].planeParams.bytesPerPix[0]);
	}

	status = NvBufSurfaceSyncForDevice(surface, -1, -1);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceSyncForDevice", status);
	status = NvBufSurfaceUnMap(surface, -1, -1);
	NVBUF_ERROR_COUT_NOTICE("NvBufSurfaceUnMap", status);
}